from io import IOBase, TextIOBase
from typing import Optional, TextIO, BinaryIO


class SSAGEBackendBase:
    def __init__(self, private_key: Optional[str] = None, public_key: Optional[str] = None):
        """
        Initialize the SSAGE Backend base object
        :param private_key: AGE private key
        :param public_key: AGE public key
        """
        if private_key is None and public_key is None:
            raise ValueError("Either a private or public key must be provided")
        if private_key is not None and public_key is not None:
            raise ValueError("Both a private and public key cannot be provided")

        self.__private_key = private_key
        self.__public_key = public_key or self.parse_public_key()

    def encrypt(self, data_in: BinaryIO, data_out: TextIO, additional_recipients: Optional[str] = None) -> None:
        """
        Encrypt the data
        :param data_in: input data
        :param data_out: output data
        :param additional_recipients: additional recipients
        """
        raise NotImplementedError()

    def decrypt(self, data_in: TextIO, data_out: BinaryIO) -> None:
        """
        Decrypt the data
        :param data_in: input data
        :param data_out: output data
        """
        raise NotImplementedError()

    def parse_public_key(self) -> str:
        """
        Parse the public key from the private key
        :return: AGE public key
        """
        raise NotImplementedError()

    @staticmethod
    def generate_private_key() -> str:
        """
        Generate a new private key
        :return: AGE private key
        """
        raise NotImplementedError()

    @property
    def private_key(self) -> Optional[str]:
        """
        Get the private key
        :return: AGE private key
        """
        return self.__private_key

    @property
    def public_key(self) -> str:
        """
        Get the public key
        :return: AGE public key
        """
        return self.__public_key



from abc import ABC
from io import StringIO, BytesIO, RawIOBase, SEEK_SET
from typing import TextIO, BinaryIO


class BytesIOPersistent(BytesIO):
    """
    A helper class to capture the data written to a BytesIO object when it is closed
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.__captured_data = None

    def close(self):
        self.__captured_data = self.getvalue()
        super().close()

    @property
    def captured_data(self):
        if not self.closed:
            return self.getvalue()

        data = self.__captured_data
        self.__captured_data = None
        return data


class StringIOPersistent(StringIO):
    """
    A helper class to capture the data written to a StringIO object when it is closed
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.__captured_data = None

    def close(self):
        self.__captured_data = self.getvalue()
        super().close()

    @property
    def captured_data(self):
        if not self.closed:
            return self.getvalue()

        data = self.__captured_data
        self.__captured_data = None
        return data


class TextIOToBinaryIOWrapper(BinaryIO, ABC):
    def __init__(self, text_stream: TextIO, encoding: str = 'ascii'):
        self.text_stream = text_stream
        self.__encoding = encoding

    def read(self, size=-1) -> bytes:
        # Read a chunk of text from the TextIO stream
        text_data = self.text_stream.read(size)
        if not text_data:  # End of the text stream
            return b""
        # Encode the text data to bytes
        return text_data.encode(self.__encoding)

    def write(self, b: bytes) -> int:
        # Decode binary data to text and write to the TextIO stream
        text_data = b.decode(self.__encoding)
        self.text_stream.write(text_data)
        return len(b)  # Return the number of bytes written

    def readable(self) -> bool:
        # Indicate that the stream is readable
        return True

    def writable(self) -> bool:
        # Indicate that the stream is writable
        return True

    def seekable(self) -> bool:
        # Indicate that the underlying stream supports seeking
        return self.text_stream.seekable()

    def seek(self, offset: int, whence: int = SEEK_SET) -> int:
        # Proxy seek to the underlying TextIO stream
        return self.text_stream.seek(offset, whence)

    def tell(self) -> int:
        # Proxy tell to the underlying TextIO stream
        return self.text_stream.tell()



import sys
from typing import Optional, TextIO, BinaryIO

from age.cli import encrypt as age_encrypt, Decryptor as AgeDecryptor, AsciiArmoredInput, AGE_PEM_LABEL
from age.keys.agekey import AgePrivateKey, AgePublicKey

from .base import SSAGEBackendBase
from .io_helpers import TextIOToBinaryIOWrapper


class SSAGEBackendAge(SSAGEBackendBase):
    def encrypt(self, data_in: BinaryIO, data_out: TextIO, additional_recipients: Optional[str] = None) -> None:
        if not hasattr(sys.stdout, 'buffer'):
            # Needed for unit tests
            sys.stdout.buffer = None

        data_out_binary = TextIOToBinaryIOWrapper(data_out)

        age_encrypt(
            recipients=[self.public_key] + (additional_recipients or []),
            infile=data_in,
            outfile=data_out_binary,
            ascii_armored=True
        )

    def decrypt(self, data_in: TextIO, data_out: BinaryIO) -> None:
        if not hasattr(sys.stdout, 'buffer'):
            # Needed for unit tests
            sys.stdout.buffer = None

        data_in_binary = AsciiArmoredInput(AGE_PEM_LABEL, TextIOToBinaryIOWrapper(data_in))

        with AgeDecryptor([self.__private_key_instance], data_in_binary) as decryptor:
            while data := decryptor.read(4096):
                data_out.write(data)

    @property
    def __private_key_instance(self) -> AgePrivateKey:
        """
        Get the private key
        :return: AGE private key
        """
        if self.private_key is None:
            raise ValueError('Private key is not available')
        return AgePrivateKey.from_private_string(self.private_key)

    @property
    def __public_key_instance(self) -> AgePublicKey:
        """
        Get the public key
        :return: AGE public key
        """
        return AgePublicKey.from_public_string(self.public_key)

    def parse_public_key(self) -> str:
        """
        Parse the public key from the private key
        :return: AGE public key
        """
        return AgePrivateKey.from_private_string(self.private_key).public_key().public_string()

    @staticmethod
    def generate_private_key() -> str:
        """
        Generate a new private key
        :return: AGE private key
        """
        return AgePrivateKey.generate().private_string()
